@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.031';
+our $VERSION = '0.032';
 
 use EV;
 use Time::HiRes;
@@ -76,6 +76,7 @@ our @EXPORT_OK = qw/
 /;
 
 our %EXPORT_TAGS = (all => \@EXPORT_OK);
+my $LAST_CONNECTION_ID;
 
 ################################################################################
 # User facing functons
@@ -83,6 +84,7 @@ our %EXPORT_TAGS = (all => \@EXPORT_OK);
 
 sub new {
     my ($class, $args) = @_;
+    $LAST_CONNECTION_ID = $$ * 1000 unless defined $LAST_CONNECTION_ID;
 
     die 'YAHC: ->new() expect args to be a hashref' if defined $args and ref($args) ne 'HASH';
     die 'YAHC: please do `my ($yahc, $yahc_storage) = YAHC::new()` and keep both these objects in the same scope' unless wantarray;
@@ -97,7 +99,6 @@ sub new {
         loop                => new EV::Loop,
         pid                 => $$, # store pid to detect forks
         storage             => \%storage,
-        last_connection_id  => $$ * 1000,
         debug               => delete $args->{debug} || $ENV{YAHC_DEBUG} || 0,
         keep_timeline       => delete $args->{keep_timeline} || $ENV{YAHC_TIMELINE} || 0,
         socket_cache        => delete $args->{socket_cache},
@@ -123,7 +124,7 @@ sub request {
     die 'YAHC: new_request() expects arguments' unless @args;
     die 'YAHC: storage object is destroyed' unless $self->{storage};
 
-    my ($conn_id, $request) = (@args == 1 ? ('connection_' . $self->{last_connection_id}++, $args[0]) : @args);
+    my ($conn_id, $request) = (@args == 1 ? ('connection_' . $LAST_CONNECTION_ID++, $args[0]) : @args);
     die "YAHC: Connection with name '$conn_id' already exists\n"
         if exists $self->{connections}{$conn_id};
 
@@ -436,15 +437,20 @@ sub _init_helper {
         _register_in_timeline($conn, "Target $scheme://$host:$port ($ip:$port) chosen for attempt #%d", $conn->{attempt})
             if exists $conn->{debug_or_timeline};
 
+        my $sock;
         my $socket_cache = $self->{socket_cache};
         my $socket_cache_id; $socket_cache_id = yahc_conn_socket_cache_id($conn) if defined $socket_cache;
-        if (defined $socket_cache_id && exists $socket_cache->{$socket_cache_id}) {
+        if (defined $socket_cache_id && (my $socket_cache_record = $socket_cache->{$socket_cache_id})) {
             _register_in_timeline($conn, "reuse socket '$socket_cache_id'") if $conn->{debug_or_timeline};
-            my $sock = $watchers->{_fh} = delete $socket_cache->{$socket_cache_id};
+            $sock = shift @{ $socket_cache_record };
+        }
+
+        if (defined $sock) {
+            $watchers->{_fh} = $sock;
             $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, sub {});
             _set_write_state($self, $conn_id);
         } else {
-            my $sock = _build_socket_and_connect($ip, $port);
+            $sock = _build_socket_and_connect($ip, $port);
             _set_connecting_state($self, $conn_id, $sock);
         }
 
@@ -866,11 +872,7 @@ sub _close_or_cache_socket {
         return;
     }
 
-    if (exists $socket_cache->{$socket_cache_id}) {
-        close(delete $socket_cache->{$socket_cache_id});
-    }
-
-    $socket_cache->{$socket_cache_id} = $fh;
+    push @{ $socket_cache->{$socket_cache_id} ||= [] }, $fh;
     _register_in_timeline($conn, "save socket '$socket_cache_id' for later use") if $conn->{debug_or_timeline};
 }
 
